﻿using StudiApp.Interfaces;
using StudiApp.Models;
using StudiApp.Services.InputValidator;
using StudiApp.Services.InputValidator.Strategy;

namespace StudiApp.Test.Unit.ServicesTests.InputValidatorTests;

internal class CourseValidatorTest
{
    private const string AllValidMessage =
        "Course Name between 1 - 100 chars\nLecturer Name between 1 - 100 chars\nEcts must be higher than 0\nLectures Name between 1 - 100 chars\nLectures StartTime has to be before EndTime\nCourse Description can't be longer than 250 chars\n";

    private const string CourseNameErrorMessage = "Course Name between 1 - 100 chars\n";
    private const string LecturerNameErrorMessage = "Lecturer Name between 1 - 100 chars\n";
    private const string EctsErrorMessage = "Ects must be higher than 0\n";
    private const string LecturesErrorMessage = "Lectures Name between 1 - 100 chars\n";
    private const string DescriptionErrorMessage = "Course Description can't be longer than 250 chars\n";
    private const string LecturesTimeErrorMessage = "Lectures StartTime has to be before EndTime\n";

    private readonly IInputValidator _inputValidator = new InputValidator();
    private Course _testCourse = new();

    [SetUp]
    public void SetUp()
    {
        _testCourse = CreateValidCourse();
    }

    private static Course CreateValidCourse()
    {
        return new Course
        {
            Name = "Course",
            LecturerName = "Lecturer",
            Ects = 4,
            Description = "Cool Course",
            Lectures = new List<Lecture>
            {
                new()
                {
                    Name = "lecture1",
                    DayTime = new DayTime
                    {
                        StartTime = new TimeOnly(10, 00),
                        EndTime = new TimeOnly(12, 00)
                    }
                },
                new()
                {
                    Name = "lecture2",
                    DayTime = new DayTime
                    {
                        StartTime = new TimeOnly(10, 00),
                        EndTime = new TimeOnly(12, 00)
                    }
                }
            }
        };
    }

    [Test]
    public void TestAllInputValid()
    {
        var result = _inputValidator.IsValid(new CourseValidatorStrategy(_testCourse));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.True);
            Assert.That(result.ErrorMessage, Is.Null);
        });
    }

    [Test]
    public void TestAllInputInvalid()
    {
        _testCourse.Name = "";
        _testCourse.LecturerName = "";
        _testCourse.Ects = 0;
        _testCourse.Lectures[0].Name = "";
        _testCourse.Lectures[0].DayTime.StartTime = new TimeOnly(14, 00);
        _testCourse.Description = new string('*', 251);

        var result = _inputValidator.IsValid(new CourseValidatorStrategy(_testCourse));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(AllValidMessage));
        });
    }

    [Test]
    public void TestCourseNameTooBig()
    {
        _testCourse.Name = new string('*', 101);

        var result = _inputValidator.IsValid(new CourseValidatorStrategy(_testCourse));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(CourseNameErrorMessage));
        });
    }

    [Test]
    public void TestCourseNameTooSmall()
    {
        _testCourse.Name = "";
        var result = _inputValidator.IsValid(new CourseValidatorStrategy(_testCourse));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(CourseNameErrorMessage));
        });
    }

    [Test]
    public void TestLecturerNameTooSmall()
    {
        _testCourse.LecturerName = "";
        var result = _inputValidator.IsValid(new CourseValidatorStrategy(_testCourse));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(LecturerNameErrorMessage));
        });
    }

    [Test]
    public void TestLecturerNameTooBig()
    {
        _testCourse.LecturerName = new string('*', 101);
        var result = _inputValidator.IsValid(new CourseValidatorStrategy(_testCourse));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(LecturerNameErrorMessage));
        });
    }

    [Test]
    public void TestEctsTooSmall()
    {
        _testCourse.Ects = 0;
        var result = _inputValidator.IsValid(new CourseValidatorStrategy(_testCourse));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(EctsErrorMessage));
        });
    }

    [Test]
    public void TestLecturesNameTooBig()
    {
        _testCourse.Lectures[0].Name = new string('*', 101);
        var result = _inputValidator.IsValid(new CourseValidatorStrategy(_testCourse));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(LecturesErrorMessage));
        });
    }

    [Test]
    public void TestLecturesTimeInvalid()
    {
        _testCourse.Lectures[0].DayTime.StartTime = new TimeOnly(14, 00);
        var result = _inputValidator.IsValid(new CourseValidatorStrategy(_testCourse));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(LecturesTimeErrorMessage));
        });
    }

    [Test]
    public void TestLecturesNameTooSmall()
    {
        _testCourse.Lectures[0].Name = "";
        var result = _inputValidator.IsValid(new CourseValidatorStrategy(_testCourse));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(LecturesErrorMessage));
        });
    }

    [Test]
    public void TestCourseDescriptionTooBig()
    {
        _testCourse.Description = new string('*', 251);
        var result = _inputValidator.IsValid(new CourseValidatorStrategy(_testCourse));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(DescriptionErrorMessage));
        });
    }

    [Test]
    public void TestCourseDescriptionNull()
    {
        _testCourse.Description = null;
        var result = _inputValidator.IsValid(new CourseValidatorStrategy(_testCourse));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.True);
            Assert.That(result.ErrorMessage, Is.Null);
        });
    }
}