﻿using StudiApp.Interfaces;
using StudiApp.Models;
using StudiApp.Services.InputValidator;
using StudiApp.Services.InputValidator.Strategy;

namespace StudiApp.Test.Unit.ServicesTests.InputValidatorTests;

public class SemesterValidatorTest
{
    private const string AllInvalidErrorMessage =
        "Semester Name between 1 - 100 chars\nSemester StartTime has to be before EndTime\n";

    private const string SemesterNameErrorMessage = "Semester Name between 1 - 100 chars\n";
    private const string SemesterDateRangeErrorMessage = "Semester StartTime has to be before EndTime\n";

    private readonly IInputValidator _inputValidator = new InputValidator();
    private Semester _testSemester = new();

    [SetUp]
    public void SetUp()
    {
        _testSemester = CreateValidSemester();
    }

    private static Semester CreateValidSemester()
    {
        return new Semester
        {
            Name = "Semester",
            DateRange = new DateRange
            {
                Start = new DateTime(2022, 01, 24),
                End = new DateTime(2022, 09, 24)
            }
        };
    }

    [Test]
    public void TestAllInputValid()
    {
        var result = _inputValidator.IsValid(new SemesterValidatorStrategy(_testSemester));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.True);
            Assert.That(result.ErrorMessage, Is.Null);
        });
    }

    [Test]
    public void TestAllInputInvalid()
    {
        _testSemester.Name = "";
        _testSemester.DateRange.End = new DateTime(2021, 01, 01);

        var result = _inputValidator.IsValid(new SemesterValidatorStrategy(_testSemester));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(AllInvalidErrorMessage));
        });
    }

    [Test]
    public void TestSemesterNameTooSmall()
    {
        _testSemester.Name = "";

        var result = _inputValidator.IsValid(new SemesterValidatorStrategy(_testSemester));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(SemesterNameErrorMessage));
        });
    }

    [Test]
    public void TestSemesterNameTooBig()
    {
        _testSemester.Name = new string('*', 101);

        var result = _inputValidator.IsValid(new SemesterValidatorStrategy(_testSemester));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(SemesterNameErrorMessage));
        });
    }

    [Test]
    public void TestSemesterDateRangeNotValid()
    {
        _testSemester.DateRange.End = new DateTime(2021, 01, 01);

        var result = _inputValidator.IsValid(new SemesterValidatorStrategy(_testSemester));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(SemesterDateRangeErrorMessage));
        });
    }
}