﻿using StudiApp.Interfaces;
using StudiApp.Models;
using StudiApp.Services.InputValidator;
using StudiApp.Services.InputValidator.Strategy;

namespace StudiApp.Test.Unit.ServicesTests.InputValidatorTests;

public class TodoValidatorTest
{
    private const string TodoNameErrorMessage = "ToDo Name between 1 - 100 chars\n";
    private const string DescriptionErrorMessage = "ToDo Description can't be longer than 250 chars\n";

    private const string TodoInvalidErrorMessage =
        "ToDo Name between 1 - 100 chars\nToDo Description can't be longer than 250 chars\n";

    private readonly IInputValidator _inputValidator = new InputValidator();
    private ToDo _testTodo = new();

    [SetUp]
    public void SetUp()
    {
        _testTodo = CreateValidTodo();
    }

    private static ToDo CreateValidTodo()
    {
        return new ToDo
        {
            Name = "ToDo"
        };
    }

    [Test]
    public void TestAllInputValid()
    {
        var result = _inputValidator.IsValid(new TodoValidatorStrategy(_testTodo));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.True);
            Assert.That(result.ErrorMessage, Is.Null);
        });
    }

    [Test]
    public void TestAllInputInvalid()
    {
        _testTodo.Name = new string('*', 101);
        _testTodo.Description = new string('*', 251);

        var result = _inputValidator.IsValid(new TodoValidatorStrategy(_testTodo));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(TodoInvalidErrorMessage));
        });
    }

    [Test]
    public void TestTodoNameTooBig()
    {
        _testTodo.Name = new string('*', 101);
        var result = _inputValidator.IsValid(new TodoValidatorStrategy(_testTodo));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(TodoNameErrorMessage));
        });
    }

    [Test]
    public void TestTodoNameTooSmall()
    {
        _testTodo.Name = "";
        var result = _inputValidator.IsValid(new TodoValidatorStrategy(_testTodo));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(TodoNameErrorMessage));
        });
    }

    [Test]
    public void TestTodoDescriptionTooBig()
    {
        _testTodo.Description = new string('*', 251);
        var result = _inputValidator.IsValid(new TodoValidatorStrategy(_testTodo));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.False);
            Assert.That(result.ErrorMessage, Is.EqualTo(DescriptionErrorMessage));
        });
    }

    [Test]
    public void TestTodoDescriptionNull()
    {
        _testTodo.Description = null;
        var result = _inputValidator.IsValid(new TodoValidatorStrategy(_testTodo));
        Assert.Multiple(() =>
        {
            Assert.That(result.IsValid, Is.True);
            Assert.That(result.ErrorMessage, Is.Null);
        });
    }
}