﻿using StudiApp.Models;

namespace StudiApp.Services.InputValidator.Strategy;

public class CourseValidatorStrategy : ValidatorStrategyBase
{
	private const string ErrorMessageCourseEcts = "Ects must be higher than 0\n";

	private const string CourseModelName = nameof(Course);
	private const string CourseLecturesName = nameof(Course.Lectures);
	private const string CourseLecturerName = "Lecturer";

	private const int EctsMinValue = 0;

	private readonly Course _course;

	public CourseValidatorStrategy(Course course)
	{
		_course = course;
	}

	public override ValidatorItem IsValid()
	{
		var courseNameValidatorItem = IsCourseNameValid();
		var courseLecturerValidatorItem = IsCourseLecturerValid();
		var courseEctsValidatorItem = IsCourseEctsValid();
		var courseLecturesNameValidatorItem = IsCourseLecturesNameValid();
		var courseLecturesTimeValidatorItem = IsCourseLecturesTimeValid();
		var courseDescriptionValidatorItem = IsCourseDescriptionValid();

		var validatorItemList = new List<ValidatorItem>
		{
			courseNameValidatorItem,
			courseLecturerValidatorItem,
			courseEctsValidatorItem,
			courseLecturesNameValidatorItem,
			courseLecturesTimeValidatorItem,
			courseDescriptionValidatorItem
		};

		return MergeValidatorItems(validatorItemList);
	}

	private ValidatorItem IsCourseLecturesTimeValid()
	{
		return _course.Lectures.All(lecture => lecture.DayTime.StartTime < lecture.DayTime.EndTime)
			? new ValidatorItem(true)
			: new ValidatorItem(false, GetErrorMessageTime(CourseLecturesName));
	}

	private ValidatorItem IsCourseDescriptionValid()
	{
		return _course.Description is null or {Length: <= DescriptionStringMaxValue}
			? new ValidatorItem(true)
			: new ValidatorItem(false, GetErrorMessageDescription(CourseModelName));
	}

	private ValidatorItem IsCourseLecturesNameValid()
	{
		return _course.Lectures.All(lecture => lecture.Name is {Length: > NameStringMinValue and <= NameStringMaxValue})
			? new ValidatorItem(true)
			: new ValidatorItem(false, GetErrorMessageName(CourseLecturesName));
	}

	private ValidatorItem IsCourseEctsValid()
	{
		return _course.Ects > EctsMinValue
			? new ValidatorItem(true)
			: new ValidatorItem(false, ErrorMessageCourseEcts);
	}

	private ValidatorItem IsCourseLecturerValid()
	{
		return _course.LecturerName is {Length: > NameStringMinValue and <= NameStringMaxValue}
			? new ValidatorItem(true)
			: new ValidatorItem(false, GetErrorMessageName(CourseLecturerName));
	}

	private ValidatorItem IsCourseNameValid()
	{
		return _course.Name is {Length: > NameStringMinValue and <= NameStringMaxValue}
			? new ValidatorItem(true)
			: new ValidatorItem(false, GetErrorMessageName(CourseModelName));
	}
}