﻿using StudiApp.Models;

namespace StudiApp.Services.InputValidator.Strategy;

public class MarkValidatorStrategy : ValidatorStrategyBase
{
	private const string MarkModelName = nameof(Mark);

	private const string ErrorMessageMarkCourse = "A course must be chosen\n";
	private const string ErrorMessageMarkValue = "Mark value out of range, must be between 1.0 and 6.0\n";
	private const string ErrorMessageMarkPercentage = "Mark Percentage out of range, must be in range 1% up to 100%\n";
	private const string ErrorMessageMarkPercentageOverall = "Percentage is above 100% over all marks of this course\n";

	private const double OverallPercentageMaxValue = 1.0;
	private const double DefaultValue = 0.0;
	private const double MarkMinValue = 1.0;
	private const double MarkMaxValue = 6.0;
	private const double PercentageMinValue = 0.0;
	private const double PercentageMaxValue = 1.0;

	private readonly Mark _mark;

	public MarkValidatorStrategy(Mark mark)
	{
		_mark = mark;
	}

	public override ValidatorItem IsValid()
	{
		var markCourseValidatorItem = IsMarkCourseValid();
		var markValueValidatorItem = IsMarkValueValid();
		var markPercentageValidatorItem = IsMarkPercentageValid();
		var markOverallPercentageValidatorItem = IsMarksOverallPercentageValid();
		var markDescriptionValidatorItem = IsMarksDescriptionValid();

		var validatorItemList = new List<ValidatorItem>
		{
			markCourseValidatorItem,
			markValueValidatorItem,
			markPercentageValidatorItem,
			markOverallPercentageValidatorItem,
			markDescriptionValidatorItem
		};

		return MergeValidatorItems(validatorItemList);
	}

	private ValidatorItem IsMarksDescriptionValid()
	{
		return _mark.Description is null or {Length: <= DescriptionStringMaxValue}
			? new ValidatorItem(true)
			: new ValidatorItem(false, GetErrorMessageDescription(MarkModelName));
	}

	private ValidatorItem IsMarksOverallPercentageValid()
	{
		return ValidatePercentageOverAllMarks()
			? new ValidatorItem(true)
			: new ValidatorItem(false, ErrorMessageMarkPercentageOverall);
	}

	private bool ValidatePercentageOverAllMarks()
	{
		
		var overallCourseMarkPercentage = _mark.Course?.Marks?.Where(m => m.Id != _mark.Id).Sum(mark => mark.Percentage) ?? DefaultValue;
		return overallCourseMarkPercentage + _mark.Percentage <= OverallPercentageMaxValue;
	}

	private ValidatorItem IsMarkPercentageValid()
	{
		return _mark.Percentage is > PercentageMinValue and <= PercentageMaxValue
			? new ValidatorItem(true)
			: new ValidatorItem(false, ErrorMessageMarkPercentage);
	}

	private ValidatorItem IsMarkValueValid()
	{
		return _mark.Value is >= MarkMinValue and <= MarkMaxValue
			? new ValidatorItem(true)
			: new ValidatorItem(false, ErrorMessageMarkValue);
	}

	private ValidatorItem IsMarkCourseValid()
	{
		return _mark.Course != null
			? new ValidatorItem(true)
			: new ValidatorItem(false, ErrorMessageMarkCourse);
	}
}