﻿using System.Collections.ObjectModel;
using CommunityToolkit.Mvvm.ComponentModel;
using CommunityToolkit.Mvvm.Input;
using Microsoft.Extensions.Logging;
using StudiApp.Interfaces;
using StudiApp.Models;
using StudiApp.Views;

namespace StudiApp.ViewModels;

public partial class CoursesOverviewViewModel : ObservableObject
{
	private const string ActiveCoursesTitleString = "Active Courses";
	private const string AllCoursesTitleString = "All Courses";

	private readonly IViewModelService _viewModelService;

	private Semester _currentSemester;

	[ObservableProperty] private ObservableCollection<Course> _courses = new();
	[ObservableProperty] private bool _isShowAllSet;
	[ObservableProperty] private ObservableCollection<Course> _shownCourses = new();
	[ObservableProperty] private string _title;

	public CoursesOverviewViewModel(IViewModelService viewModelService)
	{
		_viewModelService = viewModelService;
	}

	private ObservableCollection<Course> GetUpdatedShownCourses()
	{
		var shownCourses = new ObservableCollection<Course>();

		if (IsShowAllSet)
		{
			foreach (var course in Courses) shownCourses.Add(course);
		}
		else
		{
			var activeCourses = GetActiveCourses();

			foreach (var activeCourse in activeCourses)
			{
				shownCourses.Add(activeCourse);
			}
		}

		return shownCourses;
	}

	private IEnumerable<Course> GetActiveCourses()
	{
		return Courses.Where(course => course.Semester.Id.Equals(_currentSemester?.Id));
	}

	private string UpdateTitle()
	{
		return IsShowAllSet ? AllCoursesTitleString : ActiveCoursesTitleString;
	}

	private async Task<Semester> GetCurrentSemester()
	{
		var semesters = await _viewModelService.RepositoryManager.GetAllSemesters();
		return Semester.GetCurrent(semesters);
	}

	private async Task<ObservableCollection<Course>> GetCourses()
	{
		var courses = new ObservableCollection<Course>();

		foreach (var course in await _viewModelService.RepositoryManager.GetAllCourses()) courses.Add(course);

		return courses;
	}

	[RelayCommand]
	public void ShowAllClicked()
	{
		ShownCourses = GetUpdatedShownCourses();
		Title = UpdateTitle();
		_viewModelService.Logger.LogInformation("Show {Title}. Show only active courses: {IsShowAllSet}", Title, IsShowAllSet);
	}

	[RelayCommand]
	public async Task NewCourseClicked()
	{
		await _viewModelService.NavigationManager.NavigateTo(nameof(CoursesNewView));
		_viewModelService.Logger.LogInformation("Navigating to {ViewName} to create new Course", nameof(CoursesNewView));
	}


	[RelayCommand]
	public async Task OnAppearing()
	{
		_currentSemester = await GetCurrentSemester();

		_viewModelService.Logger.LogInformation("Current Semester loaded");

		Courses = await GetCourses();

		_viewModelService.Logger.LogInformation("All Courses have been loaded from the Database");

		ShownCourses = GetUpdatedShownCourses();

		Title = UpdateTitle();

		_viewModelService.Logger.LogInformation("{ViewModel} initialized", nameof(CoursesOverviewViewModel));
	}

	[RelayCommand]
	public async Task Tap(Course c)
	{
		await _viewModelService.NavigationManager.NavigateTo(nameof(CoursesDetailView), "Id", c.Id);
		_viewModelService.Logger.LogInformation("Navigating to {ViewName} to edit course {CourseName}", nameof(CoursesDetailView), c.Name);
	}
}