﻿using System.Collections.ObjectModel;
using CommunityToolkit.Mvvm.ComponentModel;
using CommunityToolkit.Mvvm.Input;
using Microsoft.Extensions.Logging;
using StudiApp.Interfaces;
using StudiApp.Models;
using StudiApp.Services.Converters;
using StudiApp.Services.ViewModelService;
using StudiApp.Views;

namespace StudiApp.ViewModels;
public partial class MarksOverviewViewModel : ObservableObject
{
    private readonly IViewModelService _viewModelService;

    [ObservableProperty]
    private double _totalAverageMark;

    [ObservableProperty]
    private int _totalEcts;

    [ObservableProperty]
    private double _semesterAverageMark;

    [ObservableProperty]
    private int _semesterEcts;

    [ObservableProperty] private ObservableCollection<Mark> _marks;

    [ObservableProperty] 
    private ObservableCollection<Semester> _semesters;

    [ObservableProperty] 
    private Semester _selectedSemester;

    public MarksOverviewViewModel(IViewModelService viewmodelService)
    {
        _viewModelService = viewmodelService;

        InitializeSemesterData();
    }

    [RelayCommand]
    public async void OnAppearing()
    {
        await LoadOverview();
    }

    private void InitializeSemesterData()
    {
        _viewModelService.RepositoryManager.GetAllSemesters().ContinueWith(r =>
        {
            var allSemesters = r.Result;
            Semesters = new ObservableCollection<Semester>(allSemesters);
            SelectedSemester = Semester.GetCurrent(allSemesters);
        });
    }

    private async Task LoadOverview()
    {
        var allCourses = await _viewModelService.RepositoryManager.GetAllCourses();

        if (SelectedSemester != null)
        {
            var semesterCourses = allCourses.Where(c => c.SemesterId == SelectedSemester.Id).ToList();
            var semesterMarks = GetMarksFromCourses(semesterCourses);

            Marks = new ObservableCollection<Mark>(semesterMarks);
            SemesterEcts = GetObtainedEcts(semesterCourses);
            SemesterAverageMark = AverageMarkConverter.MarksToAverage(semesterMarks);
        }

        TotalEcts = GetObtainedEcts(allCourses);
        var allMarks = await _viewModelService.RepositoryManager.GetAllMarks();
        TotalAverageMark = AverageMarkConverter.MarksToAverage(allMarks);
    }


    private static List<Mark> GetMarksFromCourses(List<Course> courses)
    {
        return courses.SelectMany(c => c.Marks).ToList();
    }

    async partial void OnSelectedSemesterChanged(Semester value)
    {
        await LoadOverview();
    }

    private static int GetObtainedEcts(IEnumerable<Course> courses)
    {
        var passedCourses = courses.Where(course => course.IsPassed);
        return passedCourses.Sum(course => course.Ects);
    }

    [RelayCommand]
    public async Task EditMark(Mark mark)
    {
	    await _viewModelService.NavigationManager.NavigateTo(nameof(MarksEditView), "Id", mark.Id);
	    _viewModelService.Logger.LogInformation($"Navigating to {nameof(MarksEditView)} (edit mark)");
    }

    [RelayCommand]
    public async Task NewMark()
    {
	    await _viewModelService.NavigationManager.NavigateTo(nameof(MarksCreateView));
        _viewModelService.Logger.LogInformation($"Navigating to {nameof(MarksCreateView)} (create new mark)");
    }

    [RelayCommand]
    public async Task DeleteMark(Mark mark)
    {
        var result = await _viewModelService.DialogService.ShowConfirmationAsync("Are you sure you want to remove this mark?");
        if (result)
        {
            await _viewModelService.RepositoryManager.DeleteMark(mark);
            Marks.Remove(mark);
            await LoadOverview();
            _viewModelService.Logger.LogInformation($"Removed mark");
        }
    }
}
